from spaceone.api.repository.v1 import plugin_pb2, plugin_pb2_grpc
from spaceone.core.pygrpc import BaseAPI


class Plugin(BaseAPI, plugin_pb2_grpc.PluginServicer):

    pb2 = plugin_pb2
    pb2_grpc = plugin_pb2_grpc

    def register(self, request, context):
        params, metadata = self.parse_request(request, context)
        with self.locator.get_service('PluginService', metadata) as plugin_svc:
            plugin_data = plugin_svc.register(params)
            return self.locator.get_info('PluginInfo', plugin_data)

    def update(self, request, context):
        params, metadata = self.parse_request(request, context)
        with self.locator.get_service('PluginService', metadata) as plugin_svc:
            plugin_data = plugin_svc.update(params)
            return self.locator.get_info('PluginInfo', plugin_data)

    def deregister(self, request, context):
        params, metadata = self.parse_request(request, context)
        with self.locator.get_service('PluginService', metadata) as plugin_svc:
            plugin_svc.deregister(params)
            return self.locator.get_info('EmptyInfo')

    def enable(self, request, context):
        params, metadata = self.parse_request(request, context)
        with self.locator.get_service('PluginService', metadata) as plugin_svc:
            plugin_data = plugin_svc.enable(params)
            return self.locator.get_info('PluginInfo', plugin_data)

    def disable(self, request, context):
        params, metadata = self.parse_request(request, context)
        with self.locator.get_service('PluginService', metadata) as plugin_svc:
            plugin_data = plugin_svc.disable(params)
            return self.locator.get_info('PluginInfo', plugin_data)

    def get_versions(self, request, context):
        params, metadata = self.parse_request(request, context)
        with self.locator.get_service('PluginService', metadata) as plugin_svc:
            version_list = plugin_svc.get_versions(params)
            return self.locator.get_info('VersionsInfo', version_list)

    def get(self, request, context):
        params, metadata = self.parse_request(request, context)
        with self.locator.get_service('PluginService', metadata) as plugin_svc:
            plugin_data = plugin_svc.get(params)
            return self.locator.get_info('PluginInfo', plugin_data)

    def list(self, request, context):
        params, metadata = self.parse_request(request, context)
        with self.locator.get_service('PluginService', metadata) as plugin_svc:
            plugins_data, total_count = plugin_svc.list(params)
            return self.locator.get_info('PluginsInfo', plugins_data, total_count, minimal=self.get_minimal(params))




from spaceone.core.pygrpc.server import GRPCServer
from .repository import Repository
from .plugin import Plugin


_all_ = ['app']

app = GRPCServer()
app.add_service(Repository)
app.add_service(Plugin)



from mongoengine import *

from spaceone.core.model.mongo_model import MongoModel

__all__ = ["Plugin"]


class Plugin(MongoModel):
    """
    name is unique per domain
    """

    plugin_id = StringField(max_length=255, unique_with="domain_id")
    name = StringField(max_length=255)
    state = StringField(
        max_length=40, default="ENABLED", choices=("ENABLED", "DISABLED")
    )
    image = StringField(max_length=255)
    registry_type = StringField(max_length=255, default="DOCKER_HUB")
    registry_config = DictField()
    resource_type = StringField(max_length=255)
    provider = StringField(max_length=255, default=None, null=True)
    capability = DictField()
    labels = ListField(StringField(max_length=255))
    tags = DictField()
    repository = DictField(default=None, null=True)
    repository_id = StringField(max_length=40)
    domain_id = StringField(max_length=255)
    created_at = DateTimeField(auto_now_add=True)
    updated_at = DateTimeField(auto_now=True)
    meta = {
        "updatable_fields": [
            "name",
            "state",
            "capability",
            "labels",
            "tags",
        ],
        "minimal_fields": [
            "plugin_id",
            "name",
            "state",
            "image",
            "registry_type",
            "resource_type",
            "provider",
        ],
        "ordering": ["name"],
        "indexes": [
            "plugin_id",
            "state",
            "registry_type",
            "resource_type",
            "provider",
            "repository_id",
            "domain_id",
        ],
    }



import copy
import logging
import jsonschema
import re

from spaceone.core.service import *
from spaceone.core import config

from spaceone.repository.error import *
from spaceone.repository.model.capability_model import Capability
from spaceone.repository.manager.plugin_manager.local_plugin_manager import (
    LocalPluginManager,
)
from spaceone.repository.manager.repository_manager import RepositoryManager

_LOGGER = logging.getLogger(__name__)

MAX_IMAGE_NAME_LENGTH = 48
REGISTRY_MAP = {
    "DOCKER_HUB": "DockerHubConnector",
    "AWS_PRIVATE_ECR": "AWSPrivateECRConnector",
    "HARBOR": "HarborConnector",
}


@authentication_handler
@authorization_handler
@mutation_handler
@event_handler
class PluginService(BaseService):
    resource = "Plugin"

    @transaction(permission="repository:Plugin.write", role_types=["DOMAIN_ADMIN"])
    @check_required(["name", "resource_type", "image", "domain_id"])
    def register(self, params):
        """Register Plugin (local repo only)

        Args:
            params (dict): {
                'name': 'str',              # required
                'resource_type': 'str',     # required
                'image': 'str',             # required
                'provider': 'str',
                'registry_type': 'str',
                'capability': 'dict',
                'labels': 'list',
                'tags': 'dict',
                'domain_id': 'str'          # injected from auth
            }

        Returns:
            plugin_info (dict)
        """

        registry_type = params.get("registry_type", "DOCKER_HUB")
        image = params["image"]
        plugin_id = self._get_image_name(image)

        # self._check_capability(params.get('capability'))
        self._check_image(image)
        self._check_registry_type(registry_type)

        plugin_mgr: LocalPluginManager = self.locator.get_manager("LocalPluginManager")

        # Only LOCAL repository can be registered
        repo_mgr: RepositoryManager = self.locator.get_manager("RepositoryManager")
        params["plugin_id"] = plugin_id
        params["repository"] = repo_mgr.get_local_repository()
        params["repository_id"] = params["repository"].get("repository_id")

        return plugin_mgr.register_plugin(params)

    @transaction(permission="repository:Plugin.write", role_types=["DOMAIN_ADMIN"])
    @check_required(["plugin_id", "domain_id"])
    def update(self, params):
        """Update Plugin. (local repo only)

        Args:
            params (dict): {
                'plugin_id': 'str',     # required
                'name': 'str',
                'capability': 'dict',
                'labels': 'list',
                'tags': 'dict'
                'domain_id': 'str'      # injected from auth
            }

        Returns:
            plugin_info (dict)
        """

        # self._check_capability(params.get('capability'))
        self._check_resource_type(params.get("resource_type"))

        plugin_mgr: LocalPluginManager = self.locator.get_manager("LocalPluginManager")
        return plugin_mgr.update_plugin(params)

    @transaction(
        permission="repository:Plugin.write",
        role_types=["DOMAIN_ADMIN"],
    )
    @check_required(["plugin_id", "domain_id"])
    def enable(self, params):
        """Enable plugin (local repo only)

        Args:
            params (dict): {
                'plugin_id': 'str',     # required
                'domain_id': 'str'      # injected from auth
            }

        Returns:
            plugin_info (dict)
        """

        plugin_mgr: LocalPluginManager = self.locator.get_manager("LocalPluginManager")
        return plugin_mgr.enable_plugin(params["plugin_id"], params["domain_id"])

    @transaction(
        permission="repository:Plugin.write",
        role_types=["DOMAIN_ADMIN"],
    )
    @check_required(["plugin_id", "domain_id"])
    def disable(self, params):
        """Disable Plugin (local repo only)

        Args:
            params (dict): {
                'plugin_id': 'str',     # required
                'domain_id': 'str'      # injected from auth
            }

        Returns:
            plugin_info (dict)
        """
        plugin_mgr: LocalPluginManager = self.locator.get_manager("LocalPluginManager")
        return plugin_mgr.disable_plugin(params["plugin_id"], params["domain_id"])

    @transaction(permission="repository:Plugin.write", role_types=["DOMAIN_ADMIN"])
    @check_required(["plugin_id", "domain_id"])
    def deregister(self, params):
        """Deregister Plugin (local repo only)

        Args:
            params (dict): {
                'plugin_id': 'str',
                'domain_id': 'str'
            }

        Returns:
            None
        """

        #########################################################################
        # Warning
        #
        # To deregister plugin, plugin has to verify there is no installed plugins
        # If there was installed plugin, other micro service can query endpoint
        # But plugin can not reply endpoint
        ##########################################################################
        # TODO: check plugin is not used

        plugin_mgr: LocalPluginManager = self.locator.get_manager("LocalPluginManager")
        return plugin_mgr.delete_plugin(params["plugin_id"], params["domain_id"])

    @transaction(
        permission="repository:Plugin.read",
        role_types=["DOMAIN_ADMIN", "WORKSPACE_OWNER", "WORKSPACE_MEMBER"],
    )
    @check_required(["plugin_id"])
    def get_versions(self, params):
        """Get Plugin version (all repositories)

        Args:
            params (dict): {
                'plugin_id': 'str',         # required
                'repository_id': 'str',
                'domain_id': 'str'          # injected from auth
            }

        Returns:
            version_list (list)
        """

        plugin_id = params["plugin_id"]
        domain_id = params.get("domain_id")
        repo_id = params.get("repository_id")

        repo_mgr: RepositoryManager = self.locator.get_manager("RepositoryManager")
        repos_info = repo_mgr.get_repositories(repo_id)

        for repo_info in repos_info:
            _LOGGER.debug(
                f"[get] find at name: {repo_info['name']}"
                f"(repo_type: {repo_info['repository_type']})"
            )
            plugin_mgr = self._get_plugin_manager_by_repo(repo_info["repository_type"])
            try:
                return plugin_mgr.get_plugin_versions(repo_info, plugin_id, domain_id)
            except Exception as e:
                _LOGGER.debug(
                    f"[get_versions] Cannot find plugin: {plugin_id} at {repo_info['name']}"
                )

        raise ERROR_NO_PLUGIN(plugin_id=plugin_id)

    @transaction(
        permission="repository:Plugin.read",
        role_types=["DOMAIN_ADMIN", "WORKSPACE_OWNER", "WORKSPACE_MEMBER", "USER"],
    )
    @check_required(["plugin_id"])
    @change_only_key({"repository_info": "repository"})
    def get(self, params):
        """Get Plugin (all repositories)

        Args:
            params (dict): {
                'query': 'dict (spaceone.api.core.v1.Query)',
                'plugin_id': 'str',
                'name': 'str',
                'state': 'str',
                'resource_type': 'str',
                'provider': 'str',
                'registry_type': 'str',
                'repository_id': 'str',
                'domain_id': 'str',        # injected from auth (optional)
            }

        Returns:
            plugin_info (dict)
        """

        plugin_id = params["plugin_id"]
        domain_id = params.get("domain_id")
        repo_id = params.get("repository_id")

        repo_mgr: RepositoryManager = self.locator.get_manager("RepositoryManager")
        repos_info = repo_mgr.get_repositories(repo_id)

        for repo_info in repos_info:
            _LOGGER.debug(
                f"[get] find at name: {repo_info['name']} "
                f"(repo_type: {repo_info['repository_type']})"
            )
            plugin_mgr = self._get_plugin_manager_by_repo(repo_info["repository_type"])
            try:
                return plugin_mgr.get_plugin(repo_info, plugin_id, domain_id)
            except Exception as e:
                _LOGGER.debug(
                    f"[get] Can not find plugin({plugin_id}) at {repo_info['name']}"
                )

        raise ERROR_NO_PLUGIN(plugin_id=plugin_id)

    @transaction(
        permission="repository:Plugin.read",
        role_types=["DOMAIN_ADMIN", "WORKSPACE_OWNER", "WORKSPACE_MEMBER", "USER"],
    )
    @change_only_key({"repository_info": "repository"}, key_path="query.only")
    @append_query_filter(
        ["plugin_id", "name", "state", "resource_type", "provider", "registry_type"]
    )
    def list(self, params):
        """List plugins (all repositories)

        Args:
            params (dict): {
                'repository_id': 'str',
                'plugin_id': 'str',
                'name': 'str',
                'state': 'str',
                'resource_type': 'str',
                'registry_type': 'str',
                'provider': 'str',
                'project_id': 'str', // deprecated
                'domain_id': 'str',
                'query': 'dict (spaceone.api.core.v1.Query)'
            }

        Returns:
            results (list): 'list of plugin_info'
            total_count (int)
        """

        query = params.get("query", {})
        query["only"] = query.get("only", [])

        if "registry_url" in query["only"]:
            query["only"].remove("registry_url")

        repo_mgr: RepositoryManager = self.locator.get_manager("RepositoryManager")
        if repository_id := params.get("repository_id"):
            repo_info = repo_mgr.get_repository(repository_id)
            plugin_mgr = self._get_plugin_manager_by_repo(repo_info["repository_type"])

            return plugin_mgr.list_plugins(repo_info, query, params)
        else:
            all_plugins_info = []
            plugin_total_count = 0
            query_page = None
            if "page" in query:
                query_page = query["page"]
                del query["page"]

            repos_info: list = repo_mgr.get_repositories()
            for repo_info in repos_info:
                plugin_mgr = self._get_plugin_manager_by_repo(
                    repo_info["repository_type"]
                )
                plugins_info, total_count = plugin_mgr.list_plugins(
                    repo_info, copy.deepcopy(query), params
                )
                all_plugins_info += plugins_info
                plugin_total_count += total_count

            if query_page:
                page_start = query_page.get("start", 1) - 1
                page_limit = query_page.get("limit")

                if page_limit:
                    all_plugins_info = all_plugins_info[
                        page_start : page_start + page_limit
                    ]

            return all_plugins_info, plugin_total_count

    def _get_plugin_manager_by_repo(self, repository_type: str):
        if repository_type == "LOCAL":
            return self.locator.get_manager("LocalPluginManager")
        elif repository_type == "MANAGED":
            return self.locator.get_manager("ManagedPluginManager")
        else:
            return self.locator.get_manager("RemotePluginManager")

    @staticmethod
    def _check_template(template):
        """
        Check json schema
        """
        if template:
            for key, value in template.items():
                if isinstance(value, dict) and "schema" in value:
                    try:
                        jsonschema.Draft7Validator.check_schema(value["schema"])
                    except Exception as e:
                        raise ERROR_INVALID_TEMPLATE_SCHEMA(key=f"template.{key}")

    @staticmethod
    def _check_capability(capability):
        if capability:
            try:
                capability_vo = Capability(capability)
                capability_vo.validate()
            except Exception as e:
                raise ERROR_INVALID_PARAMETER(key="capability", reason=e)

    @staticmethod
    def _check_service_type(name):
        """
        service_type has format rule
        format:
            <service>.<purpose>
        example:
            identity.Domain
            inventory.Collector

        Raises:
            ERROR_INVALID_PARAMETER
        """
        if name:
            pass
            # idx = name.split('.')
            # if len(idx) != 2:
            #     raise ERROR_INVALID_PARAMETER(key='service_type', reason=f'{name} format is invalid.')

    @staticmethod
    def _get_image_name(image):
        return image.rsplit("/", 1)[-1]

    @staticmethod
    def _check_registry_type(registry_type):
        registry_url = (
            config.get_global("REGISTRY_INFO", {}).get(registry_type).get("url", "")
        )

        if registry_url.strip() == "":
            raise ERROR_REGISTRY_SETTINGS(registry_type=registry_type)

    def _check_image(self, image):
        """Check image name
        format: repository/image_name
        length of image_name: < 40
        format of image_name: string and - (underscore is not allowed)
        """
        _LOGGER.debug(f"[_check_image] {image}")
        image_name = self._get_image_name(image)

        # check image_name length
        if len(image_name) > MAX_IMAGE_NAME_LENGTH:
            raise ERROR_INVALID_IMAGE_LENGTH(
                name=image_name, length=MAX_IMAGE_NAME_LENGTH
            )

        # Search naming format
        m = re.search("(?![a-zA-Z0-9\-]).*", image_name)
        if m:
            if len(m.group()) > 1:
                raise ERROR_INVALID_IMAGE_NAME_FORMAT(name=image_name)
            return True

        raise ERROR_INVALID_IMAGE_NAME_FORMAT(name=image_name)



#
#   Copyright 2020 The SpaceONE Authors.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.


import os
from setuptools import setup, find_packages

setup(
    name="spaceone-repository",
    version=os.environ.get("PACKAGE_VERSION"),
    description="SpaceONE repository service",
    long_description="",
    url="https://www.spaceone.dev/",
    author="MEGAZONE SpaceONE Team",
    author_email="admin@spaceone.dev",
    license="Apache License 2.0",
    packages=find_packages(),
    install_requires=[
        "spaceone-core",
        "spaceone-api",
        "schematics",
        "pandas",
    ],
    package_data={
        "spaceone": [
            "repository/managed_resource/plugin/*.yaml",
            "repository/managed_resource/policy/*.yaml",
            "repository/managed_resource/schema/*.yaml",
        ]
    },
    zip_safe=False,
)
